@@ -1,3 +1,5 @@
+#include "first.h"
+
 /**
  * the network chunk-API
  *
@@ -20,16 +22,27 @@
 #include <errno.h>
 #include <string.h>
 
+/* default 1MB, upper limit 128MB */
+#define DEFAULT_TEMPFILE_SIZE (1 * 1024 * 1024)
+#define MAX_TEMPFILE_SIZE (128 * 1024 * 1024)
+
+static array *chunkqueue_default_tempdirs = NULL;
+static unsigned int chunkqueue_default_tempfile_size = DEFAULT_TEMPFILE_SIZE;
+
 chunkqueue *chunkqueue_init(void) {
 	chunkqueue *cq;
 
 	cq = calloc(1, sizeof(*cq));
+	force_assert(NULL != cq);
 
 	cq->first = NULL;
 	cq->last = NULL;
 
 	cq->unused = NULL;
 
+	cq->tempdirs              = chunkqueue_default_tempdirs;
+	cq->upload_temp_file_size = chunkqueue_default_tempfile_size;
+
 	return cq;
 }
 
@@ -37,6 +50,7 @@ static chunk *chunk_init(void) {
 	chunk *c;
 
 	c = calloc(1, sizeof(*c));
+	force_assert(NULL != c);
 
 	c->type = MEM_CHUNK;
 	c->mem = buffer_init();
@@ -172,6 +186,7 @@ static void chunkqueue_prepend_chunk(chunkqueue *cq, chunk *c) {
 }
 
 static void chunkqueue_append_chunk(chunkqueue *cq, chunk *c) {
+	c->next = NULL;
 	if (cq->last) {
 		cq->last->next = c;
 	}
@@ -196,6 +211,28 @@ void chunkqueue_reset(chunkqueue *cq) {
 
 	cq->bytes_in = 0;
 	cq->bytes_out = 0;
+	cq->tempdir_idx = 0;
+}
+
+void chunkqueue_append_file_fd(chunkqueue *cq, buffer *fn, int fd, off_t offset, off_t len) {
+	chunk *c;
+
+	if (0 == len) {
+		close(fd);
+		return;
+	}
+
+	c = chunkqueue_get_unused_chunk(cq);
+
+	c->type = FILE_CHUNK;
+
+	buffer_copy_buffer(c->file.name, fn);
+	c->file.start = offset;
+	c->file.length = len;
+	c->file.fd = fd;
+	c->offset = 0;
+
+	chunkqueue_append_chunk(cq, c);
 }
 
 void chunkqueue_append_file(chunkqueue *cq, buffer *fn, off_t offset, off_t len) {
@@ -254,6 +291,24 @@ void chunkqueue_append_mem(chunkqueue *cq, const char * mem, size_t len) {
 	chunkqueue_append_chunk(cq, c);
 }
 
+
+void chunkqueue_append_chunkqueue(chunkqueue *cq, chunkqueue *src) {
+	if (src == NULL || NULL == src->first) return;
+
+	if (NULL == cq->first) {
+		cq->first = src->first;
+	} else {
+		cq->last->next = src->first;
+	}
+	cq->last = src->last;
+	cq->bytes_in += (src->bytes_in - src->bytes_out);
+
+	src->first = NULL;
+	src->last = NULL;
+	src->bytes_out = src->bytes_in;
+}
+
+
 void chunkqueue_get_memory(chunkqueue *cq, char **mem, size_t *len, size_t min_size, size_t alloc_size) {
 	static const size_t REALLOC_MAX_SIZE = 256;
 	chunk *c;
@@ -332,11 +387,25 @@ void chunkqueue_use_memory(chunkqueue *cq, size_t len) {
 	}
 }
 
+void chunkqueue_set_tempdirs_default (array *tempdirs, unsigned int upload_temp_file_size) {
+	chunkqueue_default_tempdirs = tempdirs;
+	chunkqueue_default_tempfile_size
+		= (0 == upload_temp_file_size)                ? DEFAULT_TEMPFILE_SIZE
+		: (upload_temp_file_size > MAX_TEMPFILE_SIZE) ? MAX_TEMPFILE_SIZE
+		                                              : upload_temp_file_size;
+}
+
+#if 0
 void chunkqueue_set_tempdirs(chunkqueue *cq, array *tempdirs, unsigned int upload_temp_file_size) {
 	force_assert(NULL != cq);
 	cq->tempdirs = tempdirs;
-	cq->upload_temp_file_size = upload_temp_file_size;
+	cq->upload_temp_file_size
+		= (0 == upload_temp_file_size)                ? DEFAULT_TEMPFILE_SIZE
+		: (upload_temp_file_size > MAX_TEMPFILE_SIZE) ? MAX_TEMPFILE_SIZE
+		                                              : upload_temp_file_size;
+	cq->tempdir_idx = 0;
 }
+#endif
 
 void chunkqueue_steal(chunkqueue *dest, chunkqueue *src, off_t len) {
 	while (len > 0) {
@@ -387,15 +456,13 @@ void chunkqueue_steal(chunkqueue *dest, chunkqueue *src, off_t len) {
 static chunk *chunkqueue_get_append_tempfile(chunkqueue *cq) {
 	chunk *c;
 	buffer *template = buffer_init_string("/var/tmp/lighttpd-upload-XXXXXX");
-	int fd;
+	int fd = -1;
 
 	if (cq->tempdirs && cq->tempdirs->used) {
-		size_t i;
-
 		/* we have several tempdirs, only if all of them fail we jump out */
 
-		for (i = 0; i < cq->tempdirs->used; i++) {
-			data_string *ds = (data_string *)cq->tempdirs->data[i];
+		for (errno = EIO; cq->tempdir_idx < cq->tempdirs->used; ++cq->tempdir_idx) {
+			data_string *ds = (data_string *)cq->tempdirs->data[cq->tempdir_idx];
 
 			buffer_copy_buffer(template, ds->value);
 			buffer_append_slash(template);
@@ -407,7 +474,7 @@ static chunk *chunkqueue_get_append_tempfile(chunkqueue *cq) {
 		fd = mkstemp(template->ptr);
 	}
 
-	if (-1 == fd) {
+	if (fd < 0) {
 		buffer_free(template);
 		return NULL;
 	}
@@ -426,79 +493,106 @@ static chunk *chunkqueue_get_append_tempfile(chunkqueue *cq) {
 	return c;
 }
 
-/* default 1MB, upper limit 128MB */
-#define DEFAULT_TEMPFILE_SIZE (1 * 1024 * 1024)
-#define MAX_TEMPFILE_SIZE (128 * 1024 * 1024)
+static void chunkqueue_remove_empty_chunks(chunkqueue *cq);
 
-static int chunkqueue_append_to_tempfile(server *srv, chunkqueue *dest, const char *mem, size_t len) {
-	/* copy everything to max max_tempfile_size sized tempfiles */
-	const off_t max_tempfile_size
-		= (0 == dest->upload_temp_file_size)                ? DEFAULT_TEMPFILE_SIZE
-		: (dest->upload_temp_file_size > MAX_TEMPFILE_SIZE) ? MAX_TEMPFILE_SIZE
-		                                                    : dest->upload_temp_file_size;
-	chunk *dst_c = NULL;
+int chunkqueue_append_mem_to_tempfile(server *srv, chunkqueue *dest, const char *mem, size_t len) {
+	chunk *dst_c;
 	ssize_t written;
 
-	/*
-	 * if the last chunk is
-	 * - smaller than max_tempfile_size
-	 * - not read yet (offset == 0)
-	 * -> append to it (so it might actually become larger than max_tempfile_size)
-	 * otherwise
-	 * -> create a new chunk
-	 *
-	 * */
-
-	if (NULL != dest->last
-		&& FILE_CHUNK == dest->last->type
-		&& dest->last->file.is_temp
-		&& -1 != dest->last->file.fd
-		&& 0 == dest->last->offset) {
-		/* ok, take the last chunk for our job */
-		dst_c = dest->last;
-
-		if (dest->last->file.length >= max_tempfile_size) {
-			/* the chunk is too large now, close it */
-			if (-1 != dst_c->file.fd) {
-				close(dst_c->file.fd);
-				dst_c->file.fd = -1;
-			}
-			dst_c = chunkqueue_get_append_tempfile(dest);
-		}
-	} else {
-		dst_c = chunkqueue_get_append_tempfile(dest);
-	}
-
-	if (NULL == dst_c) {
-		/* we don't have file to write to,
-		 * EACCES might be one reason.
+	do {
+		/*
+		 * if the last chunk is
+		 * - smaller than dest->upload_temp_file_size
+		 * - not read yet (offset == 0)
+		 * -> append to it (so it might actually become larger than dest->upload_temp_file_size)
+		 * otherwise
+		 * -> create a new chunk
 		 *
-		 * Instead of sending 500 we send 413 and say the request is too large
-		 */
+		 * */
 
-		log_error_write(srv, __FILE__, __LINE__, "ss",
-			"denying upload as opening temp-file for upload failed:",
-			strerror(errno));
+		dst_c = dest->last;
+		if (NULL != dst_c
+			&& FILE_CHUNK == dst_c->type
+			&& dst_c->file.is_temp
+			&& dst_c->file.fd >= 0
+			&& 0 == dst_c->offset) {
+			/* ok, take the last chunk for our job */
 
-		return -1;
-	}
+			if (dst_c->file.length >= (off_t)dest->upload_temp_file_size) {
+				/* the chunk is too large now, close it */
+				int rc = close(dst_c->file.fd);
+				dst_c->file.fd = -1;
+				if (0 != rc) {
+					log_error_write(srv, __FILE__, __LINE__, "sbss",
+						"close() temp-file", dst_c->file.name, "failed:",
+						strerror(errno));
+					return -1;
+				}
+				dst_c = NULL;
+			}
+		} else {
+			dst_c = NULL;
+		}
 
-	if (0 > (written = write(dst_c->file.fd, mem, len)) || (size_t) written != len) {
-		/* write failed for some reason ... disk full ? */
-		log_error_write(srv, __FILE__, __LINE__, "sbs",
-				"denying upload as writing to file failed:",
-				dst_c->file.name, strerror(errno));
+		if (NULL == dst_c && NULL == (dst_c = chunkqueue_get_append_tempfile(dest))) {
+			/* we don't have file to write to,
+			 * EACCES might be one reason.
+			 *
+			 * Instead of sending 500 we send 413 and say the request is too large
+			 */
 
-		close(dst_c->file.fd);
-		dst_c->file.fd = -1;
+			log_error_write(srv, __FILE__, __LINE__, "ss",
+				"opening temp-file failed:", strerror(errno));
 
-		return -1;
-	}
+			return -1;
+		}
 
-	dst_c->file.length += len;
-	dest->bytes_in += len;
+		written = write(dst_c->file.fd, mem, len);
 
-	return 0;
+		if ((size_t) written == len) {
+			dst_c->file.length += len;
+			dest->bytes_in += len;
+
+			return 0;
+		} else if (written >= 0) {
+			/*(assume EINTR if partial write and retry write();
+			 * retry write() might fail with ENOSPC if no more space on volume)*/
+			dest->bytes_in += written;
+			mem += written;
+			len -= (size_t)written;
+			dst_c->file.length += (size_t)written;
+			/* continue; retry */
+		} else if (errno == EINTR) {
+			/* continue; retry */
+		} else {
+			int retry = (errno == ENOSPC && dest->tempdirs && ++dest->tempdir_idx < dest->tempdirs->used);
+			if (!retry) {
+				log_error_write(srv, __FILE__, __LINE__, "sbs",
+						"write() temp-file", dst_c->file.name, "failed:",
+						strerror(errno));
+			}
+
+			if (0 == chunk_remaining_length(dst_c)) {
+				/*(remove empty chunk and unlink tempfile)*/
+				chunkqueue_remove_empty_chunks(dest);
+			} else {/*(close tempfile; avoid later attempts to append)*/
+				int rc = close(dst_c->file.fd);
+				dst_c->file.fd = -1;
+				if (0 != rc) {
+					log_error_write(srv, __FILE__, __LINE__, "sbss",
+						"close() temp-file", dst_c->file.name, "failed:",
+						strerror(errno));
+					return -1;
+				}
+			}
+			if (!retry) return -1;
+
+			/* continue; retry */
+		}
+
+	} while (dst_c);
+
+	return -1; /*(not reached)*/
 }
 
 int chunkqueue_steal_with_tempfiles(server *srv, chunkqueue *dest, chunkqueue *src, off_t len) {
@@ -539,7 +633,7 @@ int chunkqueue_steal_with_tempfiles(server *srv, chunkqueue *dest, chunkqueue *s
 
 		case MEM_CHUNK:
 			/* store "use" bytes from memory chunk in tempfile */
-			if (0 != chunkqueue_append_to_tempfile(srv, dest, c->mem->ptr + c->offset, use)) {
+			if (0 != chunkqueue_append_mem_to_tempfile(srv, dest, c->mem->ptr + c->offset, use)) {
 				return -1;
 			}
 
@@ -618,3 +712,19 @@ void chunkqueue_remove_finished_chunks(chunkqueue *cq) {
 		chunkqueue_push_unused_chunk(cq, c);
 	}
 }
+
+static void chunkqueue_remove_empty_chunks(chunkqueue *cq) {
+	chunk *c;
+	chunkqueue_remove_finished_chunks(cq);
+	if (chunkqueue_is_empty(cq)) return;
+
+	for (c = cq->first; c->next; c = c->next) {
+		if (0 == chunk_remaining_length(c->next)) {
+			chunk *empty = c->next;
+			c->next = empty->next;
+			if (empty == cq->last) cq->last = c;
+
+			chunkqueue_push_unused_chunk(cq, empty);
+		}
+	}
+}
