@@ -1,3 +1,5 @@
+#include "first.h"
+
 #include "plugin.h"
 #include "log.h"
 
@@ -10,7 +12,7 @@
 # include <valgrind/valgrind.h>
 #endif
 
-#ifndef __WIN32
+#if !defined(__WIN32) && !defined(LIGHTTPD_STATIC)
 # include <dlfcn.h>
 #endif
 /*
@@ -38,7 +40,6 @@ typedef enum {
 	PLUGIN_FUNC_HANDLE_SIGHUP,
 	PLUGIN_FUNC_HANDLE_SUBREQUEST,
 	PLUGIN_FUNC_HANDLE_SUBREQUEST_START,
-	PLUGIN_FUNC_HANDLE_JOBLIST,
 	PLUGIN_FUNC_HANDLE_DOCROOT,
 	PLUGIN_FUNC_HANDLE_PHYSICAL,
 	PLUGIN_FUNC_CONNECTION_RESET,
@@ -53,6 +54,7 @@ static plugin *plugin_init(void) {
 	plugin *p;
 
 	p = calloc(1, sizeof(*p));
+	force_assert(NULL != p);
 
 	return p;
 }
@@ -85,10 +87,12 @@ static int plugins_register(server *srv, plugin *p) {
 	if (0 == srv->plugins.size) {
 		srv->plugins.size = 4;
 		srv->plugins.ptr  = malloc(srv->plugins.size * sizeof(*ps));
+		force_assert(NULL != srv->plugins.ptr);
 		srv->plugins.used = 0;
 	} else if (srv->plugins.used == srv->plugins.size) {
 		srv->plugins.size += 4;
 		srv->plugins.ptr   = realloc(srv->plugins.ptr, srv->plugins.size * sizeof(*ps));
+		force_assert(NULL != srv->plugins.ptr);
 	}
 
 	ps = srv->plugins.ptr;
@@ -171,7 +175,6 @@ int plugins_load(server *srv) {
 int plugins_load(server *srv) {
 	plugin *p;
 	int (*init)(plugin *pl);
-	const char *error;
 	size_t i, j;
 
 	for (i = 0; i < srv->srvconf.modules->used; i++) {
@@ -259,8 +262,13 @@ int plugins_load(server *srv) {
 #else
 		*(void **)(&init) = dlsym(p->lib, srv->tmp_buf->ptr);
 #endif
-		if ((error = dlerror()) != NULL)  {
-			log_error_write(srv, __FILE__, __LINE__, "s", error);
+		if (NULL == init) {
+			const char *error = dlerror();
+			if (error != NULL) {
+				log_error_write(srv, __FILE__, __LINE__, "ss", "dlsym:", error);
+			} else {
+				log_error_write(srv, __FILE__, __LINE__, "ss", "dlsym symbol not found:", srv->tmp_buf->ptr);
+			}
 
 			plugin_free(p);
 			return -1;
@@ -324,7 +332,6 @@ PLUGIN_TO_SLOT(PLUGIN_FUNC_HANDLE_REQUEST_DONE, handle_request_done)
 PLUGIN_TO_SLOT(PLUGIN_FUNC_HANDLE_CONNECTION_CLOSE, handle_connection_close)
 PLUGIN_TO_SLOT(PLUGIN_FUNC_HANDLE_SUBREQUEST, handle_subrequest)
 PLUGIN_TO_SLOT(PLUGIN_FUNC_HANDLE_SUBREQUEST_START, handle_subrequest_start)
-PLUGIN_TO_SLOT(PLUGIN_FUNC_HANDLE_JOBLIST, handle_joblist)
 PLUGIN_TO_SLOT(PLUGIN_FUNC_HANDLE_DOCROOT, handle_docroot)
 PLUGIN_TO_SLOT(PLUGIN_FUNC_HANDLE_PHYSICAL, handle_physical)
 PLUGIN_TO_SLOT(PLUGIN_FUNC_CONNECTION_RESET, connection_reset)
@@ -423,6 +430,7 @@ handler_t plugins_call_init(server *srv) {
 	/* fill slots */
 
 	srv->plugin_slots = calloc(PLUGIN_FUNC_SIZEOF, sizeof(ps));
+	force_assert(NULL != srv->plugin_slots);
 
 	for (i = 0; i < srv->plugins.used; i++) {
 		size_t j;
@@ -435,6 +443,7 @@ handler_t plugins_call_init(server *srv) {
 		plugin **slot = ((plugin ***)(srv->plugin_slots))[x]; \
 		if (!slot) { \
 			slot = calloc(srv->plugins.used, sizeof(*slot));\
+			force_assert(NULL != slot); \
 			((plugin ***)(srv->plugin_slots))[x] = slot; \
 		} \
 		for (j = 0; j < srv->plugins.used; j++) { \
@@ -453,7 +462,6 @@ handler_t plugins_call_init(server *srv) {
 		PLUGIN_TO_SLOT(PLUGIN_FUNC_HANDLE_SIGHUP, handle_sighup);
 		PLUGIN_TO_SLOT(PLUGIN_FUNC_HANDLE_SUBREQUEST, handle_subrequest);
 		PLUGIN_TO_SLOT(PLUGIN_FUNC_HANDLE_SUBREQUEST_START, handle_subrequest_start);
-		PLUGIN_TO_SLOT(PLUGIN_FUNC_HANDLE_JOBLIST, handle_joblist);
 		PLUGIN_TO_SLOT(PLUGIN_FUNC_HANDLE_DOCROOT, handle_docroot);
 		PLUGIN_TO_SLOT(PLUGIN_FUNC_HANDLE_PHYSICAL, handle_physical);
 		PLUGIN_TO_SLOT(PLUGIN_FUNC_CONNECTION_RESET, connection_reset);
